// ======================================================================
//
// gl_wu.cpp
// jeff grills
//
// copyright 1996 origin systems inc
//
// ======================================================================

#include "first.h"
#include "gl_wu.h"
#include "gl_init.h"
#include "material.h"
#include "mat_wu.h"
#include "pixindex.h"
#include "palette.h"
#include "dirdraw.h"
#include "polyvert.h"
#include "clut.h"
#include "raster.h"
#include "pal_wu.h"
#include "geometry.h"
#include "gl_sfx.h"

#include <assert.h>
#include <stdio.h>

// ======================================================================

#define TRACE(a)          NOP // DPRINT_E_TRACE(a)
#define DPRINT_F(a)       NOP // DPRINT_E_JG(a)

// ======================================================================

const int RASTER_VERTEX_CNT = 16;

// ======================================================================

// -q- i would have liked to have put these in the class header, but
// doing it here means that I don't have to include david's header files
// in gl_wu.h.  Oh well, at least I made them static.

static clut_miser::color_table          colorTable;
static clut_miser::pixel_blend_table_8  blendTable;
static clut_miser::output_settings      settings;
static clut_miser::generator           *clut_generator;
static raster::vert_2d                  rasterVertex[RASTER_VERTEX_CNT];

MaterialData_wu  *GL_wu::wirefill;

// ======================================================================

#if 0
static void dump(const void *data, int length)
{
	int     i;
	const u_int8 *d8;

	d8 = reinterpret_cast<const u_int8 *>(data);
	for (i = 0; i < length; i++)
	{
		if ((i & 15) != 0)
			DPRINT_LOG((", "));
		else
			DPRINT_LOG(("  "));

		DPRINT_LOG(("0x%02x", d8[i]));

		if ((i & 15) == 15)
			DPRINT_LOG(("\n"));
	}
}
#endif

// ======================================================================

GL_wu::GL_wu(void)
{
	wirefill = NULL;
}

// ----------------------------------------------------------------------

GL_wu::~GL_wu(void)
{
	if (wirefill)
	{
		materialDestroyData(wirefill);
		wirefill = NULL;
	}

	delete clut_generator;
	clut_generator = NULL;
}

// ----------------------------------------------------------------------

void GL_wu::getInformation(Information *information)
{
	GL_abstract::getInformation(information);

	information->name         = "Software (no HSR)";
	information->versionMinor = 0;
	information->ranking      = -200;
}

// ----------------------------------------------------------------------

bool GL_wu::query(int)
{
	return true;
}

// ----------------------------------------------------------------------

bool GL_wu::init(const GL_input *input, GL_output *output)
{
	if (!GL_dd::init(input, output))
		return false;

	// create the direct draw surfaces
	DirectDraw::tripleBuffer();
	GL_dd::createSurfaces(input, output);

	// only do monochrome lighting
	Geometry::install();
	Geometry::setLightMode(Geometry::LightModeMono);

	// setup the clut generator
	settings.default_shade_bits = getShadeBits();
	settings.diffuse_scale      = 1.0f;
	settings.specular_start     = 1.0f;
	settings.specular_scale     = 1.0f;

	// setup the rasterizer color information
	switch (input->depth)
	{
		case 1:
			// built the 8bit lookup tables
			setSystemPalette(input->palette);
			raster::set_display_8(&blendTable);
		break;

		case 2:
		{
			clut_miser::pixel_format_16 pixelFormat16(DirectDraw::getRBitCount(), DirectDraw::getGBitCount(), DirectDraw::getBBitCount(), DirectDraw::getRPlaceShift(), DirectDraw::getGPlaceShift(), DirectDraw::getBPlaceShift());
			clut_generator = NEW clut_miser::generator_16(pixelFormat16);
			clut_generator->settings = settings;
			raster::set_display_16(pixelFormat16);
		}
		break;

		default:
			DEBUG_FATAL(1, ("GL_wu::init bad depth %d", input->depth));
	}


	// setup the rasterizer
	raster::set_edge_clipping(1);
	raster::set_persp_block_dim(4 << (2 - input->minor));

	// create a material to wireframe
	wirefill = NEW MaterialData_wu(255, 255, 255, 0);

	return true;
}

// ----------------------------------------------------------------------

void GL_wu::setActive(bool)
{
}

// ----------------------------------------------------------------------

void GL_wu::debugDump(void)
{
}

// ----------------------------------------------------------------------

bool GL_wu::lock(void)
{
	return GL_dd::lock() && lockDD();
}

// ----------------------------------------------------------------------

void GL_wu::unlock(bool display)
{
	GL_dd::unlock(display);
}

// ----------------------------------------------------------------------

void GL_wu::makeClut(const RGB *rgb, int numColors, byte **mem, void **clut)
{
	int                      i, size, alignMask;
	clut_miser::color_table  colorTable;

	// build color lookup table
	colorTable.num_colors = numColors;
	for (i = 0; i < numColors; i++, rgb++)
	{
		colorTable.entries[i][3] = 0;
		colorTable.entries[i][2] = rgb->r;
		colorTable.entries[i][1] = rgb->g;
		colorTable.entries[i][0] = rgb->b;
	}

	// get the size the clut will need to be
	size = clut_generator->calculate_clut_size(colorTable);

	// allocate enough memory to align the clut
	alignMask = clut_generator->suggest_clut_alignment() - 1;
	*mem = NEW byte[size + alignMask];

	// align the clut
	*clut = reinterpret_cast<void *>((reinterpret_cast<int>(*mem + alignMask)) & (~alignMask));

	// finally, make the clut
	clut_generator->make_clut(*clut, colorTable);
}

// ----------------------------------------------------------------------

void GL_wu::setSystemPalette(const Palette *pal)
{
	int                      i;
	const RGB               *rgb;
	clut_miser::generator_8 *clut_generator_8;

	DEBUG_FATAL(pal->getEntries() != 256, ("GL_wu::setColorTable need 256 colors"));

	// convert it to a color table structure
	rgb = pal->getRGB();
	colorTable.num_colors = 256;
	for (i = 0; i < 256; i++, rgb++)
	{
		colorTable.entries[i][3] = 0;
		colorTable.entries[i][2] = rgb->r;
		colorTable.entries[i][1] = rgb->g;
		colorTable.entries[i][0] = rgb->b;
	}

	// create the new clut generator for this system palette
	// -qq- we may need to invalidate all previously generated cluts
	delete clut_generator;
	clut_generator = clut_generator_8 = NEW clut_miser::generator_8(colorTable);
	clut_generator->settings = settings;
	blendTable = clut_generator_8->create_blend_table();

	// set the primary surface palette
	GL_dd::setSystemPalette(pal);
}

// ----------------------------------------------------------------------

PaletteMaterialData *GL_wu::paletteCreateMaterialData(const Palette *pal)
{
	return NEW PaletteMaterialData_wu(pal);
}

// ----------------------------------------------------------------------

void GL_wu::paletteDestroyMaterialData(PaletteMaterialData *palData)
{
	delete palData;
}

// ----------------------------------------------------------------------

MaterialData *GL_wu::materialCreateData(const StaticMaterial *material)
{
	// solid
	if (material->getRGBMap() == NULL && material->getAlphaMap() == NULL)
		return NEW MaterialData_wu(material->getR(), material->getG(), material->getB(), material->getA());

	// solid with alpha map
	if (material->getRGBMap() == NULL && material->getAlphaMap() != NULL)
		return NEW MaterialData_wu(material->getWidth(), material->getHeight(), material->getR(), material->getG(), material->getB(), material->getAlphaMap());

	// textured, optional alpha map
	return NEW MaterialData_wu(material->getRGBMap(), material->getWidth(), material->getHeight(), material->getPalette(), material->hasTransparency(), material->getAlphaMap());
}

// ----------------------------------------------------------------------

void GL_wu::materialDestroyData(MaterialData *mat)
{
	delete mat;
}

// ----------------------------------------------------------------------

void GL_wu::setCamera(const Camera &camera)
{
	Geometry::setCamera(camera);
}

// ----------------------------------------------------------------------

void GL_wu::setObjectToWorld(const Xform &objectToWorld)
{
	Geometry::setObjectToWorld(objectToWorld);
}

// ----------------------------------------------------------------------

void GL_wu::setLightList(Light **lightList, int lightCount)
{
	Geometry::setLightList(lightList, lightCount);
}

// ----------------------------------------------------------------------

bool GL_wu::pushVertexList(const Vector *vertices, int vertexCount, const Vector *normal, int normalCount)
{
	return Geometry::pushVertexList(vertices, vertexCount, normal, normalCount);
}

// ----------------------------------------------------------------------

void GL_wu::popVertexList(void)
{
	Geometry::popVertexList();
}

// ----------------------------------------------------------------------

void GL_wu::point(int pointIndex, Color color)
{
	real x, y;

	if (Geometry::point(pointIndex, &x, &y))
		GL_dd::pointNoClip(REAL_TO_INT(x), REAL_TO_INT(y), color);
}

// ----------------------------------------------------------------------

void GL_wu::line(int startIndex, int stopIndex, Color color)
{
	real x0, y0, x1, y1;

	if (Geometry::line(startIndex, stopIndex, &x0, &y0, &x1, &y1))
		GL_dd::line(REAL_TO_INT(x0), REAL_TO_INT(y0), REAL_TO_INT(x1), REAL_TO_INT(y1), color);
}

// ----------------------------------------------------------------------

void GL_wu::point(const Vector &point, Color color)
{
	real x, y;

	if (Geometry::point(point, &x, &y))
		pointNoClip(REAL_TO_INT(x), REAL_TO_INT(y), color);
}

// ----------------------------------------------------------------------

void GL_wu::line(const Vector &start, const Vector &stop, Color color)
{
	real x0, y0, x1, y1;

	if (Geometry::line(start, stop, &x0, &y0, &x1, &y1))
		GL_dd::line(REAL_TO_INT(x0), REAL_TO_INT(y0), REAL_TO_INT(x1), REAL_TO_INT(y1), color);
}

// ----------------------------------------------------------------------

void GL_wu::polygonListStart(bool, int, int, int, int)
{
	raster::set_surface(width, height, DirectDraw::screenPtr(), DirectDraw::screenPitch());
}

// ----------------------------------------------------------------------

void GL_wu::polygonListStop(void)
{
}

// ----------------------------------------------------------------------

void GL_wu::polygon(const Material *material, const PolyVertex *polyVertex, int vertexCnt, int start, bool flat)
{
	int                    i;
	const MaterialData_wu *mat;
	raster::texture       *tmap;
	const PolyVertex      *v0, *v1, *v2;
		
	DEBUG_FATAL(vertexCnt < 3 || vertexCnt > RASTER_VERTEX_CNT, ("GL_wu::polygon %d vertices", vertexCnt));

	if (DFlag(DF_E_HIDE_POLYS))
		return;

	// see if we should ignore this polygon based on the material
	mat = reinterpret_cast<const MaterialData_wu *>(material->getStaticMaterial()->getData());
	if (mat->ignoreDraw())
		return;

	// build all the raster 2d vertices, linking them counter-clockwise
	for (i = 0; i < vertexCnt; i++)
	{
		rasterVertex[i].x    = REAL_TO_FLOAT(polyVertex[i].x);
		rasterVertex[i].y    = REAL_TO_FLOAT(polyVertex[i].y);
		rasterVertex[i].next = rasterVertex+i-1;
	}
	rasterVertex[0].next   = rasterVertex+vertexCnt-1;

	// get the three points that make the largest tri
	if (start >= vertexCnt)
		start -= vertexCnt;
	v0 = polyVertex+start;
	if (start-1 < 0)
		v1 = polyVertex+(start-1+vertexCnt);
	else
		v1 = polyVertex+start-1;
	if (start-2 < 0)
		v2 = polyVertex+(start-2+vertexCnt);
	else
		v2 = polyVertex+start-2;

	if (DFlag(DF_E_WIRE_FILL))
	{
		raster::poly_outline p;

		// set the material data
		p.thickness = 1;
		p.shade     = 1;
		p.surf      = wirefill->getSolidSurf();
		p.verts     = rasterVertex+start;
		
		// draw the poly
		TRACE(("a"));
		raster::draw_poly(&p);
		TRACE(("b"));
	}
	else
	{
		// get the material information
		//mat = reinterpret_cast<const MaterialData_wu *>(material->getData());
		tmap = mat->getTmap();
		
		if (!tmap)
		{
			if (flat || DFlag(DF_E_FLAT_SHADE))
			{
				raster::poly_afs p;

				// set the polygon data
				p.surf        = mat->getSolidSurf();
				p.verts       = rasterVertex;
				p.shade       = REAL_TO_FLOAT(polyVertex[0].b);
				p.translucent = DFlag(DF_E_JEFFTEP3);

				// draw the poly
				TRACE(("c"));
				raster::draw_poly(&p);
				TRACE(("d"));
			}
			else
			{
				raster::poly_als p;

				// set the polygon data
				p.surf      = mat->getSolidSurf();
				p.verts     = rasterVertex+start;
				p.shades[0] = REAL_TO_FLOAT(v0->b);
				p.shades[1] = REAL_TO_FLOAT(v1->b);
				p.shades[2] = REAL_TO_FLOAT(v2->b);

				// draw the poly
				TRACE(("e"));
				raster::draw_poly(&p);
				TRACE(("f"));
			}
		}
		else
		{
			if (DFlag(DF_E_AFFINE))
			{
				if (flat || DFlag(DF_E_FLAT_SHADE))
				{
					raster::poly_aft p;

					// set the polygon data
					p.tmap      = tmap;
					p.verts     = rasterVertex+start;
					p.shade     = REAL_TO_FLOAT(v0->b);
					p.us[0]     = REAL_TO_FLOAT(v0->u);
					p.us[1]     = REAL_TO_FLOAT(v1->u);
					p.us[2]     = REAL_TO_FLOAT(v2->u);
					p.vs[0]     = REAL_TO_FLOAT(v0->v);
					p.vs[1]     = REAL_TO_FLOAT(v1->v);
					p.vs[2]     = REAL_TO_FLOAT(v2->v);

 					// draw the poly
					TRACE(("i"));
					raster::draw_poly(&p);
					TRACE(("j"));
				}
				else
				{
					raster::poly_alt p;

					// set the polygon data
					p.tmap      = tmap;
					p.verts     = rasterVertex+start;
					p.shades[0] = REAL_TO_FLOAT(v0->b);
					p.shades[1] = REAL_TO_FLOAT(v1->b);
					p.shades[2] = REAL_TO_FLOAT(v2->b);
					p.us[0]     = REAL_TO_FLOAT(v0->u);
					p.us[1]     = REAL_TO_FLOAT(v1->u);
					p.us[2]     = REAL_TO_FLOAT(v2->u);
					p.vs[0]     = REAL_TO_FLOAT(v0->v);
					p.vs[1]     = REAL_TO_FLOAT(v1->v);
					p.vs[2]     = REAL_TO_FLOAT(v2->v);

 					// draw the poly
					TRACE(("g"));
					raster::draw_poly(&p);
					TRACE(("h"));
				}
			}
			else
			{
				if (flat || DFlag(DF_E_FLAT_SHADE))
				{
					raster::poly_pft p;

					// set the polygon data
					p.tmap      = tmap;
					p.verts     = rasterVertex+start;
					p.shade     = REAL_TO_FLOAT(polyVertex[0].b);
					p.us[0]     = REAL_TO_FLOAT(v0->u);
					p.us[1]     = REAL_TO_FLOAT(v1->u);
					p.us[2]     = REAL_TO_FLOAT(v2->u);
					p.vs[0]     = REAL_TO_FLOAT(v0->v);
					p.vs[1]     = REAL_TO_FLOAT(v1->v);
					p.vs[2]     = REAL_TO_FLOAT(v2->v);
					p.z_invs[0] = REAL_TO_FLOAT(v0->ooz);
					p.z_invs[1] = REAL_TO_FLOAT(v1->ooz);
					p.z_invs[2] = REAL_TO_FLOAT(v2->ooz);

					TRACE(("m"));
					raster::draw_poly(&p);
					TRACE(("n"));
				}
				else
				{
					raster::poly_plt p;

					// set the polygon data
					p.tmap      = tmap;
					p.verts     = rasterVertex+start;
					p.shades[0] = REAL_TO_FLOAT(v0->b);
					p.shades[1] = REAL_TO_FLOAT(v1->b);
					p.shades[2] = REAL_TO_FLOAT(v2->b);
					p.us[0]     = REAL_TO_FLOAT(v0->u);
					p.us[1]     = REAL_TO_FLOAT(v1->u);
					p.us[2]     = REAL_TO_FLOAT(v2->u);
					p.vs[0]     = REAL_TO_FLOAT(v0->v);
					p.vs[1]     = REAL_TO_FLOAT(v1->v);
					p.vs[2]     = REAL_TO_FLOAT(v2->v);
					p.z_invs[0] = REAL_TO_FLOAT(v0->ooz);
					p.z_invs[1] = REAL_TO_FLOAT(v1->ooz);
					p.z_invs[2] = REAL_TO_FLOAT(v2->ooz);

					TRACE(("k"));
					raster::draw_poly(&p);
					TRACE(("l"));
				}
			}
		}
	}
}

// ----------------------------------------------------------------------

void GL_wu::polygon(const ScreenPoly *poly, int count)
{
	for ( ; count; --count, poly++)
		GL_wu::polygon(poly->material, poly->polyVertex, poly->numberOfVertices, 0, false);
}

// ----------------------------------------------------------------------

void GL_wu::polygon(const Poly *poly, int count)
{
	PolyVertex *polyVertex;
	int         vertexCount;

	for ( ; count; count--, poly++)
	{
		if (Geometry::polygon(poly, !DFlag(DF_E_NO_3D_CLIP), &polyVertex, &vertexCount))
		{
			int start;

			if (vertexCount > 3)
			{
				int i, j, k;
				real size, largestSize;

				// -q- okay, for all you people that hate my for loops, this is a sick and twisted one with two, count them, two back references!
				// find the largest triangle in the quad, and set start to be the last vertex in that tri
				for (k = vertexCount-2, j = vertexCount-1, i = 0, start = 0, largestSize = 0; i < vertexCount; k = j, j = i, i++)
				{
					size =
						(polyVertex[i].x * polyVertex[k].y) - (polyVertex[i].y * polyVertex[k].x) +
						(polyVertex[k].x * polyVertex[j].y) - (polyVertex[k].y * polyVertex[j].x) +
						(polyVertex[j].x * polyVertex[i].y) - (polyVertex[j].y * polyVertex[i].x);

					if (i == 0 || size > largestSize)
					{
						start       = i;
						largestSize = size;
					}
				}
			}
			else
				start = 0;

			GL_wu::polygon(poly->material, polyVertex, vertexCount, start, (poly->flags & (PF_FLAT_SHADE | PF_FULLY_LIT)) != 0);
		}
	}
}

// ----------------------------------------------------------------------

bool GL_wu::specialEffect(int id, void *data)
{
//	switch (id)
//	{
//		case GL_SFX_LENS_FLARE:
			/* put your code here too, jason */
//			return true;
//	}

	return GL_dd::specialEffect(id, data);
}

// ======================================================================
